/*
 * Copyright: (C) 1999 Bruce W. Forsberg
 *
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2.1 of the License, or any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Lesser General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public
 *   License along with this library; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
 *
 *   Bruce Forsberg  forsberg@adnc.com
 *
 */


#include <stdio.h>
#include <stdlib.h>

#include <deque.h>

#include "aflibAudio.h"
#include "aflibData.h"


aflibAudio::aflibAudio() : aflibChain()
{
   // This node is enabled by default
   _enable = TRUE;
}

aflibAudio::aflibAudio(
   const aflibAudio& audio) : aflibChain(audio)
{
   // This node is enabled by default
   _enable = TRUE;

   setAudioConfig(audio.getAudioConfig());
}


aflibAudio::~aflibAudio()
{
}

void
aflibAudio::enable(bool enable)
{
   // This allows this node to be enabled or disabled. This assists in removing
   // a node from a chain without having to destruct it. This is useful to enable
   // an application to toggle a filter in and out for example.

   _enable = enable;
}

bool
aflibAudio::getEnable() const
{
   return(_enable);
}

void
aflibAudio::setAudioConfig(const aflibConfig& cfg)
{
   _cfg = cfg;
}

const aflibConfig&
aflibAudio::getAudioConfig() const
{
   return (_cfg);
}


aflibData *
aflibAudio::process(
   aflibStatus& ret_status,
   long long position,
   int num_samples,
   bool free_memory) const
{
   // NOTE: if 0 for num_samples then we pick. 
   // This will only handle one parent for this child. If the derived class has more than
   // one parent then it will need to override this function and call what ever process
   // functions in its parents it needs. If the caller passes FALSE to free_memory then
   // the pointer to the afData class containing the data will be returned. It is then the
   // responsibility of the calling app to free this memory using delete. If no parameter
   // is passed then it will not be returned by default. Instead NULL will be returned.

   int  list_size = 0;
   aflibData * data;

   ret_status = AFLIB_SUCCESS;

   list<const aflibAudio *> audio_list = this->getParents();
   list_size = audio_list.size();

   if (list_size > 1)
   {
      cerr << "aflibAudio::process SOFTWARE ERROR! More than one parent not allowed" << endl;
      exit(1);
   }

   // IF no more parents then process since we are at the end of the chain
   if (list_size == 0)
   {
      // IF node is not enabled then skip processing
      if (_enable == TRUE)
      {
         if (num_samples == 0)
            data = new aflibData(4096);
         else
            data = new aflibData(num_samples);

         ret_status = compute_segment(*data, position);
      }
   }
   // ELSE call parent and let it process it first then process result
   else
   {
      data = (*audio_list.begin())->process(ret_status, position, num_samples, FALSE);
      if (ret_status == AFLIB_SUCCESS)
      {
         // ptr can be NULL if parent was not enabled
         if (data == NULL)
         {
            if (num_samples == 0)
               data = new aflibData(4096);
            else
               data = new aflibData(num_samples);
         }

         if (_enable == TRUE)
         {
            ret_status = compute_segment(*data, position);
         }
      }
   }

   // IF caller does not want memory returned then free
   if (free_memory == TRUE)
   {
      delete data;
      data = NULL;
   }

   return (data);
}


