/*
 * Copyright: (C) 1999 Bruce W. Forsberg
 *
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2.1 of the License, or any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Lesser General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public
 *   License along with this library; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
 *
 *   Bruce Forsberg  forsberg@adnc.com
 *
 */


#ifndef _AFLIBAUDIOEDIT_H_
#define _AFLIBAUDIOEDIT_H_

#include "list.h"
#include "set.h"

#include "aflibAudio.h"
#include "aflibAudioClip.h"

// This is a class that simplifies audio editing. This is derived from the aflibAudio base
// class so it can be used in an audio chain. This is a complex class that aids the 
// application programmer in writing an audio editing program. It can accept multiple
// inputs and allows the user to add audio clip segments, remove audio clip segments, 
// and add or remove inputs. One can use this object in a chain so that the output of
// this object can be output to an audio file or device for output. Once the clip
// information is stored in this class it will automatically get the audio input from
// the correct input and route it to the output. 
// One limitation of this object is that all inputs must have the same number of
// channels, samples / second, and data size.
// We also provide our own process function instead of using the base classes since
// we need to retrieve data from different inputs.
//
// ADDING AND REMOVING INPUTS
// One can add or remove inputs to this class. The constructor does require one input.
// An ID number is returned when additional inputs are added. the first input specified
// should be assumed to be 1. When a input is removed any audio clips that are obtained
// from this input will be removed. If all inputs are removed then FALSE will be sent
// to enable() on the aflibAudio base class to disable this object from the chain. When a
// input is added again this object will be enabled. The functions to add and remove
// inputs are:
//   int  addInput(const aflibAudio& input);
//   void removeInput(int input);
//
// ADDING AND REMOVING AUDIO SEGMENTS
// Adding and removing segments is the heart of this object. This in effect allows one
// to map inputs to the output of this object. A typical application would be to have
// only one input and then add a segment that contains the entire audio segment of the
// input. Then the user can remove certain audio segments of this input. This makes
// the output look like one continuous audio stream without the segments that have
// been removed. The APIs here allow one to specify segments in either samples or
// segments. If one uses doubles then the specifcation is assumed to be seconds and if
// long long then samples. One can also remove a segment by the segment number. One
// should verify the segment number first thought as segment numbers can change after
// segments are added, removed, or inputs removed. The functions to add and remove
// segments are:
//   void addSegment(
//      int  input,
//      long long input_start_position,
//      long long input_stop_position,
//      long long output_insert_position);
//   void addSegment(
//      int  input,
//      double input_start_seconds,
//      double input_stop_seconds,
//      double output_insert_seconds);
//   void removeSegment(int seg_num);
//   void removeSegment(
//      long long output_start_position,
//      long long output_stop_position);
//   void removeSegment(
//      double output_start_seconds,
//      double output_stop_seconds);
//
// OBTAINING INFORMATION
// Several functions exist to get information on the audio segment data. One can get the
// total number of current segments and then one can obtain the data for any particular
// segment in either seconds or samples. Segments start at number 1.
//   int getNumberOfSegments();
//   void getSegment(
//      int segment_number,
//      int&  input,
//      long long& input_start_position,
//      long long& input_stop_position,
//      long long& output_start_position,
//      long long& output_stop_position);
//   void getSegment(
//      int segment_number,
//      int&  input,
//      double& input_start_seconds,
//      double& input_stop_seconds,
//      double& output_start_seconds,
//      double& output_stop_seconds);
// UNDO / REDO
// TBD. In the future Undo / Redo capability will be added to undo the last audio segment
// addition or deletion.
//


class aflibAudioEdit : public aflibAudio {

public:

   // Available contructors and destructors
   aflibAudioEdit(const aflibAudio& audio);
   ~aflibAudioEdit();

   // Add and remove inputs
   int
   addInput(const aflibAudio& input);

   void
   removeInput(int input);

   // Add and remove segments
   void
   addSegment(
      int  input,
      long long input_start_position,
      long long input_stop_position,
      long long output_insert_position);

   void
   addSegment(
      int  input,
      double input_start_seconds,
      double input_stop_seconds,
      double output_insert_seconds);

   void
   removeSegment(int seg_num);

   void
   removeSegment(
      long long output_start_position,
      long long output_stop_position);

   void
   removeSegment(
      double output_start_seconds,
      double output_stop_seconds);

   // Obtain information
   int
   getNumberOfSegments();

   void
   getSegment(
      int segment_number,
      int&  input,
      long long& input_start_position,
      long long& input_stop_position,
      long long& output_start_position,
      long long& output_stop_position);

   void
   getSegment(
      int segment_number,
      int&  input,
      double& input_start_seconds,
      double& input_stop_seconds,
      double& output_start_seconds,
      double& output_stop_seconds);

   // Perform Undo / Redo of an audio edit
   aflibUndoRedo
   getUndoRedoStatus() const;

   void
   performUndoRedo();

   aflibData *
   process(
      aflibStatus& ret_status,
      long long position,
      int num_samples,
      bool free_memory = TRUE) const;

   aflibStatus
   compute_segment(
      aflibData& data,
      long long position = -1) ;

   // Callback from afChain
   void
   parentWasDestroyed(int parent_id);

private:

   aflibAudioEdit();

   aflibAudioEdit(const aflibAudioEdit& op);

   const aflibAudioEdit&
   operator=(const aflibAudioEdit& op);

   void
   printClips();

   void
   recomputeConfig();

list<int>  _input_array;
set< aflibAudioClip, less < aflibAudioClip > > _clip_array;

};


#endif
