/*
 * Copyright: (C) 1999 Bruce W. Forsberg
 *
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2.1 of the License, or any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Lesser General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public
 *   License along with this library; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
 *
 *   Bruce Forsberg  forsberg@adnc.com
 *
 */


#include <stdio.h>
#include <iostream.h>
#include <time.h>
#include <sys/stat.h>


#include "aflibAudioRecorder.h"
#include "aflibAudioFile.h"
#include "aflibData.h"
#include "aflibConfig.h"


aflibAudioRecorder::aflibAudioRecorder(const aflibAudio& audio) :
   aflibAudio(audio)
{
}

aflibAudioRecorder::~aflibAudioRecorder()
{
}

void
aflibAudioRecorder::addRecordItem(
   const aflibDateTime& start_date,
   const aflibDateTime& stop_date,
   const string& file,
   aflibFileType file_type,
   long long max_limit,
   long long each_limit,
   aflibConfig& config)
{
   aflibRecorderItem audio(
      start_date, stop_date, file, file_type, max_limit, each_limit, config);

   // Add new record item to list
   _item_list.push_back(audio);
}

void
aflibAudioRecorder::removeRecordItem(
   int item)
{
   list<aflibRecorderItem>::iterator it;
   int i = 0;

   // Iterate thru list and till we find the item
   for (it = _item_list.begin(); it != _item_list.end(); it++, i++)
   {
      // When we find the item remove
      if (i == item)
      {
         _item_list.erase(it);
         break;
      }
   }
}

int
aflibAudioRecorder::getNumberOfRecordItems() const
{
   int total_size;

   // Return size of list as number of items
   total_size = _item_list.size();

   return(total_size);
}

void
aflibAudioRecorder::getItem(
   int item,
   aflibDateTime& start_date,
   aflibDateTime& stop_date,
   string& file, 
   aflibFileType& file_type, 
   long long& max_limit,
   long long& each_limit,
   aflibConfig& config) const
{
   list<aflibRecorderItem>::const_iterator it;
   int i = 0;

   // Iterate until we get to the correct item
   for (it = _item_list.begin(); it != _item_list.end(); it++, i++)
   {
      // When we find the item return the results
      if (i == item)
      {
         start_date = (*it).getStartDate();
         stop_date = (*it).getStopDate();
         file = (*it).getAudioFile();
         file_type = (*it).getAudioFileType();
         max_limit = (*it).getMaxFileLimit();
         each_limit = (*it).getEachFileLimit();
         config = (*it).getConfig();
         break;
      }
   }
}


aflibStatus
aflibAudioRecorder::compute_segment(
   aflibData& data,
   long long position) 
{
   aflibDateTime current_time;
   list<aflibRecorderItem>::iterator it;
   aflibStatus status = AFLIB_SUCCESS;
   aflibAudioFile * audio_file = NULL;
   aflibConfig config;

   current_time = getCurrentTime();

   for (it = _item_list.begin(); it != _item_list.end(); it++)
   {
      // We are done
      if ((*it).getStopDate() <= current_time)
      {
         // If we are currently recording then stop
         if ((*it).getFileObject())
         {
            delete (*it).getFileObject();
            (*it).setFileObject(NULL);

         }
      }
      // We have not started yet
      else if (current_time < (*it).getStartDate())
      {
         // Nothing to do
      }
      // We are currently recording
      else
      {
         // Are we currently writing to a file
         if ((*it).getFileObject())
         {
            // Are we done? Have we reached the file limit.
            if (audioFileSizeCheck((*it)) == FALSE)
            {
               // Set stop date to indicate stop time
               (*it).setStopDate(current_time);
            }
         }
         // ELSE create a file
         else
         {
            config = data.getConfig();
            audio_file = new aflibAudioFile(*this, (*it).getAudioFileType(),
               (*it).getAudioFile(), &config, &status);

            (*it).setFileObject(audio_file);
         }
         // Write out a data segment
         (*it).getFileObject()->compute_segment(data, position);
      }
   }

   return (AFLIB_SUCCESS);
}


aflibDateTime
aflibAudioRecorder::getCurrentTime()
{
   aflibDateTime  cur_time;
   time_t unix_time;
   struct tm tm_time;

   // Get the current date and time
   time(&unix_time);
   localtime_r(&unix_time, &tm_time);

   // Set all the fields in our date time structure
   cur_time.setYear(tm_time.tm_year + 1900);
   cur_time.setMonth(tm_time.tm_mon + 1);
   cur_time.setDay(tm_time.tm_mday + 1);
   cur_time.setHour(tm_time.tm_hour);
   cur_time.setMinute(tm_time.tm_min);
   cur_time.setSecond(tm_time.tm_sec);

   return(cur_time);
}

bool
aflibAudioRecorder::audioFileSizeCheck(aflibRecorderItem& item)
{
   const char * file = NULL;
   struct stat  statbuf;
   bool ret_result = TRUE;
   aflibConfig config;
   aflibStatus status;
   aflibAudioFile * audio_file = NULL;

   file = item.getAudioFile().c_str();

   stat(file, &statbuf);

   // Is each file size implemented
   if (item.getEachFileLimit() != -1)
   {
      //  Has the total length of an audio file been exceeded
      if (statbuf.st_size > item.getEachFileLimit())
      {
         config = item.getFileObject()->getConfig();

         // close old file and free memory
         delete item.getFileObject();

         // Get new file name and create new file object and store
         item.processNextFile();

         audio_file = new aflibAudioFile(*this, item.getAudioFileType(),
            item.getAudioFile(), &config, &status);
         item.setFileObject(audio_file);
        
         // Store this file size as total
         item.setCurrentSize( item.getCurrentSize() + statbuf.st_size );
      }
   }

   // Is total file size implemented
   if (item.getMaxFileLimit() != -1)
   {
      ret_result = FALSE;
   }
   return (ret_result);
}


