/*
 * Copyright: (C) 1999 Bruce W. Forsberg
 *
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2.1 of the License, or any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Lesser General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public
 *   License along with this library; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
 *
 *   Bruce Forsberg  forsberg@adnc.com
 *
 */


#include <stdio.h>
#include <iostream.h>

#include "aflibData.h"


aflibData::aflibData(long length)
{
   // _data is not initialized yet
   _initialized = FALSE;
   _data = NULL;
   _length = length;
   _total_length = 0;
   _byte_inc = 0;
}

aflibData::aflibData(
   const aflibConfig& config,
   long length)
{
   _config = config;
   _length = length;

   _byte_inc = _config.getBitsPerSample() / 8;

   init();
}

aflibData::aflibData(const aflibData& data)
{
   _config = data.getConfig();
   data.getLength(_length);
   _byte_inc = _config.getBitsPerSample() / 8;

   init();

   if (data.getDataPointer() != NULL)
   {
      memcpy(_data, data.getDataPointer(), _total_length);   
   }
}

aflibData&
aflibData::operator=(const aflibData& data)
{
   _config = data.getConfig();
   data.getLength(_length);
   _byte_inc = _config.getBitsPerSample() / 8;

   init();

   if (data.getDataPointer() != NULL)
   {
      memcpy(_data, data.getDataPointer(), _total_length);   
   }
   return (*this);
}

aflibData::~aflibData()
{
   delete [] _data;
}

void
aflibData::setConfig(const aflibConfig& config)
{
   _config = config;

   _byte_inc = _config.getBitsPerSample() / 8;

   init();
}

const aflibConfig&
aflibData::getConfig() const
{
   return (_config);
}

void
aflibData::getLength(long& length) const
{
   // This returns the number of samples. It does not matter how many channels.
   // It only means samples no matter how many channels.

   length = _length;
}

long
aflibData::getTotalLength()
{
   // This returns the total length of memory that is allocated.

   return _total_length;
}

void *
aflibData::getDataPointer() const
{
   return _data;
}


void
aflibData::setSample(
   int sample,
   long position,
   int channel)
{
   // First position begins at 0
   // First channel begins at 0
   // This function needs a much more efficient algoritm and needs to be
   // an inline function if at all possible.
   // We also need to handle endianness.


   // Need to add support for other data layouts
   if (_config.getDataOrientation() == AFLIB_INTERLEAVE_LE)
   {
      long data_position = _config.getChannels() * position +
                           channel;
#if 0
      if ((data_position * _byte_inc + _byte_inc) > _total_length)
      {
         cerr << "Buffer Overrun! ERROR!" << endl;
         return;
      }
#endif

      if (_config.getSampleSize() == AFLIB_DATA_16S)
      {
         signed short int * ptr = (signed short int *)_data;
         ptr[data_position] = (signed short int)sample;
      }
      else if (_config.getSampleSize() == AFLIB_DATA_8U)
      {
         unsigned char * ptr = (unsigned char *)_data;
         ptr[data_position] = (unsigned char)sample;
      }
      else if (_config.getSampleSize() == AFLIB_DATA_8S)
      {
         signed char * ptr = (signed char *)_data;
         ptr[data_position] = (signed char)sample;
      }
      else if (_config.getSampleSize() == AFLIB_DATA_16U)
      {
         unsigned short int * ptr = (unsigned short int *)_data;
         ptr[data_position] = (unsigned short int)sample;
      }
   }
}


int
aflibData::getSample(
   long position,
   int channel)
{
   int ret_value = 0;


   // Need to add support for other data layouts
   if (_config.getDataOrientation() == AFLIB_INTERLEAVE_LE)
   {
      long data_position = _config.getChannels() * position +
                           channel;
#if 0
      if ((data_position * _byte_inc + _byte_inc) > _total_length)
      {
         cerr << "Buffer Overrun! ERROR!" << endl;
         return 0;
      }
#endif

      if (_config.getSampleSize() == AFLIB_DATA_16S)
      {
         signed short int * ptr = (signed short int *)_data;
         ret_value = (int) ptr[data_position];
      }
      else if (_config.getSampleSize() == AFLIB_DATA_8U)
      {
         unsigned char * ptr = (unsigned char *)_data;
         ret_value = (int) ptr[data_position];
      }
      else if (_config.getSampleSize() == AFLIB_DATA_8S)
      {
         signed char * ptr = (signed char *)_data;
         ret_value = (int) ptr[data_position];
      }
      else if (_config.getSampleSize() == AFLIB_DATA_16U)
      {
         unsigned short int * ptr = (unsigned short int *)_data;
         ret_value = (int)ptr[data_position];
      }
   }

   return (ret_value);
}

void
aflibData::getMinMax(
   int& min_value,
   int& max_value) const
{
   // This returns the minimum and maximum values of the data that is
   // stored.

   switch (_config.getSampleSize())
   {
      case AFLIB_DATA_8S:
         min_value = -127;
         max_value = 128;
      break;

      case AFLIB_DATA_8U:
         min_value = 0;
         max_value = 255;
      break;

      case AFLIB_DATA_16S:
         min_value = -32767;
         max_value = 32768;
      break;

      case AFLIB_DATA_16U:
         min_value = 0;
         max_value = 65536;
      break;

      default:
         min_value = 0;
         max_value = 0;
      break;
   }
}


void
aflibData::init()
{
   delete [] _data;
   _data = NULL;

   _total_length = _config.getChannels() * _length * _byte_inc;
   _data = new unsigned char [_total_length];
   if (_data == NULL)
   {
      cerr << "Error. Out of memory" << endl;
   }
   else
   {
      _initialized = TRUE;
   }
}



