/*
 * Copyright: (C) 1999 Bruce W. Forsberg
 *
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2.1 of the License, or any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Lesser General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public
 *   License along with this library; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
 *
 *   Bruce Forsberg  forsberg@adnc.com
 *
 */

// Base class for audio file reading and writing

#ifndef _AFLIBFILE_H
#define _AFLIBFILE_H

#include "aflib.h"
#include "aflibConfig.h"
#include "bstring.h"
#include <stdio.h>
#include <stdlib.h>

// This is the base class for all audio file and device objects. It provides
// the functionality for calling derived classes and defines the API for these
// derived classes. The main entry APIs are the open and create static
// functions. Users actually use these. This will in turn allocate an object
// of the proper derived class.
//
// This class supports the ability to dynamically load and free these derived
// objects. These derived objects will each be made into a library that will
// be loaded only when needed. This area needs much work. In order to call
// the derived API in a shared object one needs to call with the mangaled name.
// These are all hardcoded for now but a better way needs to be found and 
// added. All OS's are most likely going to be a little bit different. This
// is needed though since this makes this library light weight and very
// powerfull. 
//
// Also in the future we need a method to add new file and device libraries
// without changing the code. Such as a configuration file that could be set
// at startup and all file formats registered.

class aflibData;

class aflibFile {

public:

   virtual ~aflibFile();

   static aflibFile *
   open(
      aflibFileType type_enum,
      const string& file,
      aflibConfig* cfg = NULL,
      aflibStatus* status = NULL);

   static aflibFile *
   create(
      aflibFileType type_enum,
      const string& file,
      const aflibConfig& cfg,
      aflibStatus* status = NULL);

   virtual aflibStatus
   afopen(
      const string& file,
      aflibConfig* cfg);

   virtual aflibStatus
   afcreate(
      const string& file,
      const aflibConfig& cfg);

   virtual aflibStatus
   afread(
      aflibData& data,
      long long position = -1);

   virtual aflibStatus
   afwrite(
      aflibData& data,
      long long position = -1);

   // These next "is" functions define for a class whether the class will support data sent to it
   // in this format. These do not define whether a file writes to this format. These are used so
   // that if data needs to be converted to a class this will tell us to what formats to go to.
   // Another words this helps us to determine the interface to a class not what the class does.
   virtual bool
   isSizeSupported(
      aflib_data_size data_size);

   virtual bool
   isEndianSupported(
      aflib_data_endian data_end);

   const aflibConfig&
   getConfig() const;

protected:

   aflibFile();

aflibConfig   _cfg;
long       _total_size;
long       _one_second_size;
long       _current_place_size;

void *     _lib1;
char       _lib_name[256];

/* These are used by derived classes */
FILE *     _fd_FILE;
int        _fd_int;

private:

   aflibFile(int num);

   aflibFile(const aflibFile& file);

   const aflibFile&
   operator== (const aflibFile& file);

};


#endif
