/*
 * Copyright: (C) 1999 Bruce W. Forsberg
 *
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2.1 of the License, or any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Lesser General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public
 *   License along with this library; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
 *
 *   Bruce Forsberg  forsberg@adnc.com
 *
 */

// class for Device audio file reading and writing

#include <stdio.h>
#include <stdlib.h>
#include <iostream.h>
#include <sys/ioctl.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/soundcard.h>

#include "aflibDevFile.h"
#include "aflibData.h"


aflibDevFile::aflibDevFile()
{
   _fd_int = -1;
}

aflibDevFile::~aflibDevFile()
{
   if (_fd_int != -1)
   {
      close(_fd_int);
   }
}

aflibStatus
aflibDevFile::afopen(
   const string& file,
   aflibConfig* cfg)
{
   aflibStatus status = AFLIB_SUCCESS;
   int snd_format;
   int snd_stereo;
   int snd_speed;
   const char * file_string;

   file_string = file.c_str();

   if ((_fd_int = ::open(file_string, O_RDONLY, 0)) == -1)
   {
      cout << "Unable to open device" << endl;
      return (AFLIB_ERROR_OPEN);
   }

   // Store defaults if user has specified
   if (cfg != NULL)
   {
      _cfg = *cfg;
   }

   if (cfg != NULL)
   {
      if (cfg->getBitsPerSample() == 8)
      {
         snd_format = AFMT_U8;
      }
      else if (cfg->getBitsPerSample() == 16)
      {
         snd_format = AFMT_S16_LE;
      }
      else
      {
         // Some unknown setting
         snd_format = AFMT_U8;
      }
   }
   else
   {
      snd_format = AFMT_S16_LE;
   }
   _cfg.setDataOrientation(AFLIB_INTERLEAVE_LE);

   if (snd_format == AFMT_S16_LE)
   {
      _cfg.setSampleSize(AFLIB_DATA_16S);
   }
   else
   {
      _cfg.setSampleSize(AFLIB_DATA_8U);
   }

   // We need to change the buffering in driver so that we get somewhat real time
   // performance. This provides 2 buffers of 8096 bytes each. In future may need
   // to be based on config passed in.
   int arg = 0x0002000D;
   if (ioctl(_fd_int, SNDCTL_DSP_SETFRAGMENT, &arg) == -1)
   {
      cout << "Unable to set buffer sizes" << endl;
   }
      
   if (ioctl(_fd_int, SNDCTL_DSP_SETFMT, &snd_format) == -1)
   {
      cout << "Unable to program format in device" << endl;
      close (_fd_int);
      _fd_int = -1;
      return (AFLIB_ERROR_UNSUPPORTED);
   }

   if (cfg != NULL)
   {
      _cfg.setChannels(cfg->getChannels());
   }
   else
   {
      _cfg.setChannels(1);
   }
   snd_stereo = _cfg.getChannels() - 1;
   if (ioctl(_fd_int, SNDCTL_DSP_STEREO, &snd_stereo) == -1)
   {
      cout << "Unable to program channels in device" << endl;
      close (_fd_int);
      _fd_int = -1;
      return (AFLIB_ERROR_UNSUPPORTED);
   }

   if (cfg != NULL)
   {
      snd_speed = cfg->getSamplesPerSecond();
   }
   else
   {
      snd_speed = 44100;
   }
   _cfg.setSamplesPerSecond(snd_speed);
   if (ioctl(_fd_int, SNDCTL_DSP_SPEED, &snd_speed) == -1)
   {
      cout << "Unable to program speed in device" << endl;
      close (_fd_int);
      _fd_int = -1;
      return (AFLIB_ERROR_UNSUPPORTED);
   }

   int frag_size = 0;
   ioctl(_fd_int, SNDCTL_DSP_GETBLKSIZE, &frag_size) ;
   cout << "Frag Size is " << frag_size << endl;

   return (status);
}

aflibStatus
aflibDevFile::afcreate(
   const string& file,
   const aflibConfig& config)
{
   int snd_format;
   int snd_stereo;
   int snd_speed;
   int mask = 0;
   aflibStatus status = AFLIB_SUCCESS;

   if ((_fd_int = ::open(file.c_str(), O_WRONLY, 0)) == -1)
   {
      cout << "Unable to open device" << endl;
      return (AFLIB_ERROR_OPEN);
   }

   // We need to change the buffering in driver so that we get somewhat real time
   // performance. This provides 2 buffers of 8096 bytes each. In future may need
   // to be based on config passed in. 
   int arg = 0x0002000D;
   if (ioctl(_fd_int, SNDCTL_DSP_SETFRAGMENT, &arg) == -1)
   {
      cout << "Unable to set buffer sizes" << endl;
   }
   
   if (ioctl(_fd_int, SNDCTL_DSP_GETFMTS, &mask) == -1)
   {
      cout << "Unable to get supported formats" << endl;
   }

   if (mask & AFMT_MU_LAW)
   {
      cout << "AFMT_MU_LAW is supported" << endl;
   }

   if (mask & AFMT_A_LAW)
   {
      cout << "AFMT_A_LAW is supported" << endl;
   }

   if (mask & AFMT_IMA_ADPCM)
   {
      cout << "AFMT_MU_ADPCM is supported" << endl;
   }

   if (mask & AFMT_U8)
   {
      cout << "AFMT_U8 is supported" << endl;
   }

   if (mask & AFMT_S16_LE)
   {
      cout << "AFMT_S16_LE is supported" << endl;
   }

   if (mask & AFMT_S16_BE)
   {
      cout << "AFMT_S16_BE is supported" << endl;
   }

   if (mask & AFMT_S8)
   {
      cout << "AFMT_S8 is supported" << endl;
   }

   if (mask & AFMT_U16_LE)
   {
      cout << "AFMT_U16_LE is supported" << endl;
   }

   if (mask & AFMT_U16_BE)
   {
      cout << "AFMT_U16_BE is supported" << endl;
   }

   if (mask & AFMT_MPEG)
   {
      cout << "AFMT_MPEG is supported" << endl;
   }

   if ((config.getBitsPerSample() == 16) && (config.getDataOrientation() == AFLIB_INTERLEAVE_LE))
   {
      snd_format = AFMT_S16_LE;
   }
   else if (config.getBitsPerSample() == 8)
   {
      snd_format = AFMT_U8;
   }
   else
   {
      cout << "Unsupported sample format" << endl;
      close (_fd_int);
      _fd_int = -1;
      return (AFLIB_ERROR_UNSUPPORTED);
   }

   if (ioctl(_fd_int, SNDCTL_DSP_SETFMT, &snd_format) == -1)
   {
      cout << "Unable to program format in device" << endl;
      close (_fd_int);
      _fd_int = -1;
      return (AFLIB_ERROR_UNSUPPORTED);
   }

   snd_stereo = config.getChannels() - 1;
   if (ioctl(_fd_int, SNDCTL_DSP_STEREO, &snd_stereo) == -1)
   {
      cout << "Unable to program channels in device" << endl;
      close (_fd_int);
      _fd_int = -1;
      return (AFLIB_ERROR_UNSUPPORTED);
   }

   snd_speed = config.getSamplesPerSecond();   
   if (ioctl(_fd_int, SNDCTL_DSP_SPEED, &snd_speed) == -1)
   {
      cout << "Unable to program speed in device" << endl;
      close (_fd_int);
      _fd_int = -1;
      return (AFLIB_ERROR_UNSUPPORTED);
   }

   int frag_size = 0;
   ioctl(_fd_int, SNDCTL_DSP_GETBLKSIZE, &frag_size) ;
   cout << "Frag Size is " << frag_size << endl;

   return(status);
}


aflibStatus
aflibDevFile::afread(
   aflibData& data,
   long long position )
{
   void * p_data;
   long   total_length;
   aflibStatus  status = AFLIB_SUCCESS;

   data.setConfig(_cfg);
   total_length = data.getTotalLength();
   p_data = data.getDataPointer();

   read(_fd_int, p_data, total_length);

   return (status);
}

aflibStatus
aflibDevFile::afwrite(
   aflibData& data,
   long long position )
{
   void * p_data;
   long   total_length;
   aflibStatus  status = AFLIB_SUCCESS;

   total_length = data.getTotalLength();
   p_data = data.getDataPointer();

   if (write(_fd_int, p_data, total_length) == -1)
   {
      perror("");
      cerr << "Failed to write to audio device " << endl;
   }
   return (status);
}


